/****************************************************************************
** $Id: gltexobj.cpp,v 1.7 2002/02/10 16:28:25 tim Exp tim $
**
** Copyright (C) 1992-2000 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

/****************************************************************************
**
** This is a simple QGLWidget demonstrating the use of QImages for textures.
**
** Much of the GL code is inspired by the 'spectex' and 'texcyl' 
** public domain demo programs by Brian Paul
**
** This has been hacked somewhat by Tim Colgate to support drawing a tree.
** The background, rotation and fog is set up here, but all the rest of the
** drawing code is in the tree classes.
****************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <iostream.h>

#include "gltexobj.h"
#include <qimage.h>
#include <qtimer.h>
#include <GL/glu.h>

#include "tree.h"
#include "tree1.h"	// until plugins work (qq)


const int redrawWait = 50;
const GLfloat fogColor[4] = { 0.8f, 0.8f, 0.8f, 1.0f };

/*!
  Create a GLTexobj widget
*/

GLTexobj::GLTexobj(QWidget* parent, const char* name)
    : QGLWidget(parent, name), tree(0)
{
	xRot = yRot = zRot = 0.0;		// default object rotation
	rotoff = 0;
	xrotdir = true;
	scale = 2.1;			// default object scale
	object = 0;
	rotating = false;
	zooming = false;
	zoom = true;
	growing = false;
	grow = true;
	show_leaves = true;
	autumn = false;
	season = 0;
	wind = false;
	fog = false;

	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), SLOT(update()));
	timer->start(redrawWait, TRUE);
}


/*!
  Release allocated resources
*/

GLTexobj::~GLTexobj()
{
	delete tree;
}

void GLTexobj::draw_sky()
{

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);	// disable lighting effects on background

	glBindTexture(GL_TEXTURE_2D, texnum_sky);

	float z=300.0;
	float x=200;
	float y=100;
	glBegin(GL_QUADS);
	// The sky is drawn on two parts, a vertical quad in the distance and a horzontal quad high up.

	/* Bottom Left Of The Texture and Quad */
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-x, -0.0f, -z);
	/* Bottom Right Of The Texture and Quad */
	glTexCoord2f(0.0f, 0.0f); glVertex3f(x, -0.0f, -z);
	/* Top Right Of The Texture and Quad */
	glTexCoord2f(0.0f, 0.5f); glVertex3f(x,  y, -z);
	/* Top Left Of The Texture and Quad */
	glTexCoord2f(1.0f, 0.5f); glVertex3f(-x,  y, -z);

	/* Bottom Left Of The Texture and Quad */
	glTexCoord2f(1.0f, 0.5f); glVertex3f(-x, y, -z);
	/* Bottom Right Of The Texture and Quad */
	glTexCoord2f(0.0f, 0.5f); glVertex3f(x, y, -z);
	/* Top Right Of The Texture and Quad */
	glTexCoord2f(0.0f, 1.0f); glVertex3f(x,  y, 130);
	/* Top Left Of The Texture and Quad */
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-x,  y, 130);

	glEnd();
}

void GLTexobj::draw_grass()
{
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

	glBindTexture(GL_TEXTURE_2D, texnum_grass);

	float z=300.0;
	float x=200;
	float y=0;
	glBegin(GL_QUADS);

	// The "grass" texture is drawn horizontally.
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-x, y, 50);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(x, y, 50);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(x,  y, -z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-x,  y, -z);

	glEnd();
}

/*!
  Paint the texobj. The actual openGL commands for drawing the texobj are
  performed here.
*/
void GLTexobj::paintGL()
{
	// if rotating, tree will spin (unrealistically) on the spot, and the viewpoint
	// will move up and around the tree.
	if (rotating) {
		if (xrotdir) {
			rotoff += 1.0;
			if (rotoff >= 100)
				xrotdir = !xrotdir;
		}
		else {		
			rotoff -= 1.0;
			if (rotoff <= 0)
				xrotdir = !xrotdir;
	}    			
		
	//    xRot += 1.0;
	//    yRot += 2.5;
	yRot += 5.0;
	}

	// zooming draws the tree larger or smaller. It doesn't change the viewpoint.
	if (zooming) {
		if (zoom)
			if (scale >= 5)
				zoom = !zoom;
			else
				scale += 0.1;

		else
			if (scale < 0.2)
				zoom = !zoom;
			else
				scale -= 0.1;
	}

	// Draw the tree as it would have looked when it was younger.
	if (growing) {
		if (grow)
			if (age >= max_age)
				grow = !grow;
			else
				age += 0.025;
		else
			if (age <= 0.5)
				grow = !grow;
			else
				age -= 0.025;
	}

	// At present this just changes the leaf colour from greens to browns.
	// For most conifers, this doesn't make a lot of sense!
	if (season != 0) {
		season--;
		if (autumn)
			tree->setColour(season/100.0);
		else
			tree->setColour(1 - season/100.0);
		
		if (!animating())
			timer->stop();
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glPushMatrix();
	gluLookAt(rotoff/10.0-5, rotoff/1.2+4, 70  -rotoff/2.0,	// eye pos
		 0, 25, 0,		// center
		 0, 1, 0);			// up is +ve Y
	draw_sky();
	draw_grass();
	glRotatef(xRot, 1.0, 0.0, 0.0);
	glRotatef(yRot, 0.0, 0.1, 0.0);
	glRotatef(zRot, 0.0, 0.0, 1.0);
	glScalef(scale, scale, scale);

	tree->draw_tree(age);

	glPopMatrix();

	if (animating()) {
		glFlush(); // Make sure everything is drawn before restarting timer
		timer->start(redrawWait, TRUE); // Wait this many msecs before redraw
	}
}


/*!
  Set up the OpenGL rendering state, and define display list
*/
void GLTexobj::initializeGL()
{
	// Set up the lights
	GLfloat whiteDir[4] = {2.0, 2.0, 2.0, 1.0};
	GLfloat whiteAmb[4] = {1.0, 1.0, 1.0, 1.0};
	GLfloat lightPos[4] = {30.0, 30.0, 30.0, 1.0};


	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, whiteAmb);

	glMaterialfv(GL_FRONT, GL_DIFFUSE, whiteDir);
	glMaterialfv(GL_FRONT, GL_SPECULAR, whiteDir);
	glMaterialf(GL_FRONT, GL_SHININESS, 20.0);

	glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteDir);		// enable diffuse
	glLightfv(GL_LIGHT0, GL_SPECULAR, whiteDir);	// enable specular
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

	// fog
	//GLuint fogMode[] = { GL_EXP, GL_EXP2, GL_LINEAR };
	// GL_EXP2: 0.019, 115, 220
	glFogi(GL_FOG_MODE, GL_LINEAR); /* Fog Mode */
	glFogfv(GL_FOG_COLOR, fogColor);         /* Set Fog Color */
	glFogf(GL_FOG_DENSITY, 1.9f);           /* How Dense Will The Fog Be */
	//    glHint(GL_FOG_HINT, GL_DONT_CARE);       /* Fog Hint Value */
	glFogf(GL_FOG_START, 45.0f);              /* Fog Start Depth */
	glFogf(GL_FOG_END, 75.0f);                /* Fog End Depth */

	// texture
	// Set up the background textures
	QImage buf, tex1;

	if (!buf.load("grass.bmp")) {   // Load first image from file
	cerr << "Could not read image file grass.bmp";
	exit(1);
	}
	tex1 = QGLWidget::convertToGLFormat(buf);  // flipped 32bit RGBA
	glGenTextures(1, &texnum_grass);
	glBindTexture(GL_TEXTURE_2D, texnum_grass);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, tex1.width(), tex1.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, tex1.bits());

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	if (!buf.load("sky.bmp")) {   // Load first image from file
	cerr << "Could not read image file sky.bmp.";
	exit(1);
	}
	tex1 = QGLWidget::convertToGLFormat(buf);  // flipped 32bit RGBA
	glGenTextures(1, &texnum_sky);
	glBindTexture(GL_TEXTURE_2D, texnum_sky);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, tex1.width(), tex1.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, tex1.bits());

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glEnable(GL_TEXTURE_2D);

	// Set up various other stuff

	glClearColor(0.0, 0.0, 0.0, 0.0); // Let OpenGL clear to black
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);  	
// 	glEnable(GL_NORMALIZE);
// 	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	// now make the tree
	age = max_age = 12;
	tree = new Tree1(max_age, 2.0);	// age, length
	tree->setLeaves(show_leaves);
}


/*!
  Set up the OpenGL view port, matrix mode, etc.
*/
void GLTexobj::resizeGL(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// void glFrustum(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near_val, GLdouble far_val);
	glFrustum(-2.0, 2.0, -2.3, 2.5, 5.0, 400.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, -25.0, -60.0);
}


/*!
  Set the rotation angle of the object to \e degrees around the X axis.
*/
void GLTexobj::setXRotation(int degrees)
{
	xRot = (GLfloat)(degrees % 360);
	updateGL();
}


/*!
  Set the rotation angle of the object to \e degrees around the Y axis.
*/
void GLTexobj::setYRotation(int degrees)
{
	yRot = (GLfloat)(degrees % 360);
	updateGL();
}


/*!
  Set the rotation angle of the object to \e degrees around the Z axis.
*/
void GLTexobj::setZRotation(int degrees)
{
	zRot = (GLfloat)(degrees % 360);
	updateGL();
}


/*!
  Turns animation on or off
*/
void GLTexobj::toggleRotating()
{
	rotating = !rotating;
	if (rotating)
		updateGL();
	else
		if (!animating())
			timer->stop();
}

void GLTexobj::toggleZooming()
{
	zooming = !zooming;
	if (zooming)
		updateGL();
	else
		if (!animating())
			timer->stop();
}

void GLTexobj::toggleGrowing()
{
	growing = !growing;
	if (growing)
		updateGL();
	else
		if (!animating())
			timer->stop();
}

void GLTexobj::toggleLeaves()
{
	show_leaves = !show_leaves;
	tree->setLeaves(show_leaves);
	updateGL();
}

void GLTexobj::toggleAutumn()
{
	autumn = !autumn;
	season = 100-season;
	updateGL();
}

void GLTexobj::toggleWind()
{
	wind = !wind;
	tree->setWind(wind);
	updateGL();
}

void GLTexobj::toggleFog()
{
	fog = !fog;
	if (fog) {
		glEnable(GL_FOG);
		glClearColor(fogColor[0], fogColor[1], fogColor[2], 1);
	}
	else {
		glDisable(GL_FOG);
		glClearColor(0.0, 0.0, 0.0, 0.0); // Let OpenGL clear to black
	}
	updateGL();
}

void GLTexobj::newTree()
{
	delete tree;
	tree = new Tree1(10, 2.0);	// age, length
	tree->setLeaves(show_leaves);
	updateGL();
}
